#!/usr/bin/env python
# -*- python -*-

import optparse
import os.path
import re
import sys
import urllib

import pygtk
pygtk.require("2.0")
import gconf
import gobject
import gtk
import gtk.glade

import pygst
pygst.require("0.10")
import gst

NAME = "AudioFormat"
VERSION = "0.2"

try:
    import nautilus
    HAS_NAUTILUS = True
except ImportError:
    print >>sys.stderr, "No Nautilus extension found, disabling Nautilus menu modification"
    HAS_NAUTILUS = False

def find_glade_file(name):
    FORMATS = ("%s/share/%s.glade", "%s/share/audioformat/%s.glade",
               "%s/audioformat/%s.glade")
    class PathWalker:
        def __init__(self, path):
            self._base = [path]
            self._updirs = []
            self._path = None
        def get_path(self):
            if self._path is None:
                self._path = os.path.realpath(
                    apply(os.path.join, self._base + self._updirs))
            return self._path
        def higher(self):
            self._updirs.append("..")
            self._path = None
    pw = PathWalker(os.path.dirname(__file__))
    path = pw.get_path()
    while path != "/":
        path = pw.get_path()
        for format in FORMATS:
            f = format % (path, name)
            if os.path.exists(f):
                return f
        pw.higher()
        newpath = pw.get_path()
        if newpath == path:
            break
        path = newpath
    raise Exception("Don't know where to find glade file %s" % name)

class ProfileError(ValueError):
    pass

class AudioProfile(gobject.GObject, object):
    """Models an audio profile in GConf."""
    def __init__(self, name, description, pipeline, extension, active):
        gobject.GObject.__init__(self)
        def empty(v):
            return v is None or str(v).strip() == ""
        def anyempty(*values):
            return True in (empty(v) for v in values) 
        if anyempty(name, pipeline, extension):
            raise ProfileError("No values may be empty or undefined")
        self._name = name
        self._description = description
        self._pipe = pipeline
        self._ext = extension
        # always True or False
        if active:
            self._active = True
        else:
            self._active = False
        return

    name = property(lambda self: self._name, None, None, 
                    "Name of the profile")
    description = property(lambda self: self._description, 
                           None, None, "Description of the profile")
    pipeline = property(lambda self: self._pipe,
                        None, None, "The GStreamer pipeline")
    extension = property(lambda self: self._ext,
                         None, None, "The file extension")
    active = property(lambda self: self._active,
                      None, None, "Is the profile active")

class AudioProfiles(gobject.GObject):
    """Handles all the audio profiles in GConf."""
    GCONF_ROOT = "/system/gstreamer/0.10/audio"
    GCONF_DIR = "%s/profiles" % GCONF_ROOT
    GCONF_PROFILE_LIST = "%s/global/profile_list" % GCONF_ROOT

    SIGNAL_AUDIO_PROFILES_UPDATED = "audio-profiles-updated"

    __gsignals__ = {
        SIGNAL_AUDIO_PROFILES_UPDATED : 
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())
        }

    def __init__(self):
        gobject.GObject.__init__(self)
        self._client = gconf.client_get_default()
        self._client.add_dir(self.GCONF_DIR, gconf.CLIENT_PRELOAD_RECURSIVE)
        self._client.connect("value_changed", self.audio_profiles_changed_cb)
        self._client.notify_add(self.GCONF_PROFILE_LIST, self.profile_list_changed_cb)
        self._profiles = {}
        return

    def read_audio_profiles(self):
        for profile_key in self.read_audio_profiles_list():
            try:
                if not self._profiles.has_key(profile_key):
                    profile = AudioProfile(
                        *[self.read_profile_value(profile_key, n) for n in [
                                "name", "description", "pipeline",
                                "extension", "active"]])
                    self._profiles[profile_key] = profile
                    yield profile
                else:
                    yield self._profiles[profile_key]
            except ProfileError, e:
                pass
        return

    profiles = property(read_audio_profiles, None, None, 
                        "Generate the list of profiles")

    # private

    def read_audio_profiles_list(self):
        return self._client.get_list(self.GCONF_PROFILE_LIST, gconf.VALUE_STRING)
    
    def read_profile_value(self, profile, value):
        v = self._client.get("%s/%s/%s" % (self.GCONF_DIR, profile, value))
        if v is None:
            return None
        if v.type == gconf.VALUE_STRING:
            return v.get_string()
        elif v.type == gconf.VALUE_BOOL:
            return v.get_bool()
        return v

    def audio_profiles_changed_cb(self, client, item, data):
        self.invalidate_profile(item)
        self.emit(self.SIGNAL_AUDIO_PROFILES_UPDATED)
        return

    def profile_list_changed_cb(self, client, notify_id, entry, *args):
        self._profiles = {}
        self.emit(self.SIGNAL_AUDIO_PROFILES_UPDATED)
        return

    def invalidate_profile(self, key):
        tail = self.key_tail(key)
        if self._profiles.has_key(tail[0]):
            del self._profiles[tail[0]]
        return

    def key_tail(self, key):
        return key[len(self.GCONF_DIR)+1:].split("/")

class ConversionSettingsView(gobject.GObject):
    PROFILE_COMBO = "profile_combo"
    EDIT_BUTTON = "edit_profiles_button"
    DESCRIPTION_LABEL = "profile_description_label"
    FILECHOOSER = "conversion_target_chooser"

    SIGNAL_PROFILE_SELECTED = "profile-selected"
    SIGNAL_DIRECTORY_CHANGED = "directory-changed"

    __gsignals__ = {
        SIGNAL_PROFILE_SELECTED : 
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_OBJECT,)),
        SIGNAL_DIRECTORY_CHANGED :
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_STRING,))
        }

    def __init__(self, tree):
        gobject.GObject.__init__(self)
        self._profile_combo = self.initialize_combo(
            tree.get_widget(self.PROFILE_COMBO))
        self._edit_button = tree.get_widget(self.EDIT_BUTTON)
        self._description_label = tree.get_widget(self.DESCRIPTION_LABEL)
        self._filechooser = tree.get_widget(self.FILECHOOSER)
        self._audio_profiles = AudioProfiles()
        self._audio_profiles.connect(
            AudioProfiles.SIGNAL_AUDIO_PROFILES_UPDATED, self.profiles_updated)
        self._profile_combo.set_model(self.create_model())
        tree.signal_autoconnect(self)
        return

    def get_directory(self):
        return self._filechooser.get_current_folder()

    # private

    def initialize_combo(self, widget):
        renderer = gtk.CellRendererText()
        widget.pack_start(renderer)
        widget.set_attributes(renderer, text=0)
        return widget

    def create_model(self):
        model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_OBJECT)
        for profile in self._audio_profiles.profiles:
            s = profile.name
            model.append((s, profile))
        return model

    def profiles_updated(self, *args):
        self._profile_combo.set_model(self.create_model())
        return

    def on_profile_changed(self, widget, *data):
        iter = widget.get_active_iter()
        model = widget.get_model()
        ob = model.get(iter, 1)[0]
        self._description_label.set_text(ob.description)
        self.emit(self.SIGNAL_PROFILE_SELECTED, ob)
        return

    def on_target_directory_changed(self, widget, *data):
        dir = widget.get_filename()
        self.emit(self.SIGNAL_DIRECTORY_CHANGED, dir)
        return

    def on_edit_profiles_button_clicked(self, button, *args):
        os.spawnlp(os.P_NOWAIT, "gnome-audio-profiles-properties")
        return

class ConvertProgressView:
    TREE_NAME = "tracks_treeview"
    COLUMN_NAME = 0
    COLUMN_PROGRESS = 1
    COLUMN_CURRENT = 2
    COLUMN_OBJECT = 3

    def __init__(self, tree):
        self._treeview = self.initialize_tree(tree.get_widget(self.TREE_NAME))
        return

    def initialize_tree(self, widget):
        model = gtk.ListStore(str, int, bool, gobject.TYPE_OBJECT)
        widget.set_model(model)
        
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn("File name", renderer, 
                                    text=self.COLUMN_NAME)
        widget.append_column(column)

        renderer = gtk.CellRendererProgress()
        column = gtk.TreeViewColumn("Progress", renderer, 
                                    value=self.COLUMN_PROGRESS)
        widget.append_column(column)
        
        return widget

    def set_tracks(self, tracks):
        model = self._treeview.get_model()
        for track in tracks:
            model.append((track.source.rsplit("/", 1)[-1], 0, False, track))
        return

    def track_finished(self, track):
        for line in self._treeview.get_model():
            if line[self.COLUMN_OBJECT] is track:
                line[self.COLUMN_PROGRESS] = 100
        return

    def set_current_track(self, track):
        for line in self._treeview.get_model():
            if line[self.COLUMN_CURRENT] and line[self.COLUMN_OBJECT] is not track:
                line[self.COLUMN_CURRENT] = False
            elif line[self.COLUMN_OBJECT] is track:
                line[self.COLUMN_CURRENT] = True
        return
    
    def set_current_position(self, position):
        for line in self._treeview.get_model():
            if line[self.COLUMN_CURRENT]:
                line[self.COLUMN_PROGRESS] = position
        return

class AudioConverterDialog:
    GLADE_FILE = "convert_progress_dialog"
    DIALOG_NAME = "convert_dialog"
    EXECUTE_BUTTON = "conversion_execute_button"
    NOTEBOOK_NAME = "dialog_notebook"

    def __init__(self, audio_converter, own_mainloop):
        self._audio_converter = audio_converter
        self._audio_converter.connect(
            AudioConverter.SIGNAL_TRACK_STARTED, self.track_started_cb)
        self._audio_converter.connect(
            AudioConverter.SIGNAL_TRACK_COMPLETED, self.track_completed_cb)
        glade_tree = gtk.glade.XML(find_glade_file(self.GLADE_FILE))

        self._dialog = glade_tree.get_widget(self.DIALOG_NAME)
        self._execute_button = glade_tree.get_widget(self.EXECUTE_BUTTON)
        self._notebook = glade_tree.get_widget(self.NOTEBOOK_NAME)

        self._profile_selection_view = ConversionSettingsView(glade_tree)
        self._convert_progress_view = ConvertProgressView(glade_tree)

        self._profile_selection_view.connect(
            ConversionSettingsView.SIGNAL_PROFILE_SELECTED,
            self.profile_selected_cb)
        self._profile_selection_view.connect(
            ConversionSettingsView.SIGNAL_DIRECTORY_CHANGED,
            self.directory_changed_cb)

        self._current_profile = None
        self._current_directory = None

        glade_tree.signal_autoconnect(self)
        self._dialog.show()

        self._ticker = None
        self._own_mainloop = own_mainloop
        return

    # private

    def profile_selected_cb(self, ob, profile):
        self._current_profile = profile
        self._execute_button.set_sensitive(True)
        return

    def directory_changed_cb(self, ob, directory):
        self._current_directory = directory
        return

    def on_conversion_execute_button_clicked(self, button, *args):
        self._notebook.set_current_page(1)
        self._execute_button.hide()
        if self._current_directory is None:
            self._current_directory = (
                self._profile_selection_view.get_directory())
        self._audio_converter.target_directory = self._current_directory
        self._audio_converter.target_profile = self._current_profile
        self._audio_converter.prime_for_conversion()
        self._convert_progress_view.set_tracks(self._audio_converter.tracks)
        self._audio_converter.convert_tracks()
        return

    def exit(self, destroy):
        if self._own_mainloop:
            gtk.main_quit()
        elif destroy:
            self._dialog.destroy()
        return
            
    def on_delete_event(self, event, *args):
        self.exit(False)
        return

    def on_destroy_event(self, event, *args):
        self.exit(False)
        return

    def on_cancel_clicked(self, button, *args):
        self.exit(True)
        return

    def track_started_cb(self, ac, track, *args):
        self._convert_progress_view.set_current_track(track)
        gobject.timeout_add(500, self.update_position_cb, track)
        return

    def track_completed_cb(self, ac, track, *args):
        self._convert_progress_view.track_finished(track)
        return

    def update_position_cb(self, track):
        if track.complete:
            return False
        else:
            position = track.get_position()
            self._convert_progress_view.set_current_position(position)
        return True

class AudioConverter(gobject.GObject, object):
    __slots__ = ('_track_names', '_tracks', '_target_directory', 
                 '_target_profile', '_convert_queue')

    SIGNAL_TRACK_COMPLETED = "completion"
    SIGNAL_TRACK_STARTED = "started"

    __gsignals__ = {
        SIGNAL_TRACK_COMPLETED :
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_OBJECT,)),
        SIGNAL_TRACK_STARTED :
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_OBJECT,))
        }

    def __init__(self):
        gobject.GObject.__init__(self)
        self._track_names = []
        self._tracks = []
        self._target_directory = None
        self._target_profile = None
        self._convert_queue = []
        return

    def get_target_directory(self):
        return self._target_directory

    def set_target_directory(self, target_directory):
        self._target_directory = target_directory
        return
    target_directory = property(
        get_target_directory, set_target_directory, None,
        "Target directory for the conversion")

    def get_track_names(self):
        return (tn for tn in self._track_names)

    def set_track_names(self, tracks):
        self._track_names = tracks
        return
    track_names = property(get_track_names, set_track_names, None,
                           "Names of the tracks to convert")

    def get_target_profile(self):
        return self._target_profile

    def set_target_profile(self, target_profile):
        self._target_profile = target_profile
        return
    target_profile = property(get_target_profile, set_target_profile, None,
                             "Target profile for the conversion")

    def get_tracks(self):
        return [t for t in self._tracks]

    def set_tracks(self, tracks):
        self._tracks = tracks
        return
    tracks = property(get_tracks, set_tracks, None, "The Track objects")

    def get_track_directory(self, track_name, common_prefix, target_directory):
        tn_middle = os.path.dirname(track_name)[len(common_prefix):]
        return os.path.join(target_directory, tn_middle)

    def prime_for_conversion(self):
        if self._target_directory is None or self._target_profile is None:
            raise Exception("Target directory (%s) or profile (%s) not set" % 
                            (self._target_directory, self._target_profile))
        common_prefix = os.path.commonprefix(self._track_names)
        self._tracks = [
            Track(t, self.target_profile, 
                  self.get_track_directory(t, common_prefix, 
                                           self.target_directory))
            for t in self._track_names]
        self._convert_queue = [t for t in self._tracks]
        self._convert_queue.reverse()
        return

    def convert_tracks(self):
        if len(self._convert_queue) > 0:
            self.convert_first()
        return

    def convert_first(self):
        first = self._convert_queue.pop()
        first.connect(Track.SIGNAL_TRACK_COMPLETED, self.completed_cb)
        first.convert()
        self.emit(self.SIGNAL_TRACK_STARTED, first)
        return

    def completed_cb(self, track, *args):
        self.emit(self.SIGNAL_TRACK_COMPLETED, track)
        self.convert_tracks()
        return

class Pipeline(gobject.GObject, object):
    """A wrapper for a GStreamer pipeline containing the whole
    src ! encode ! sink chain."""
    __slots__ = ('_filesrc', '_encoder', '_filesink', '_gst_pipe')

    SIGNAL_TRACK_COMPLETED = "completion"

    __gsignals__ = {
        SIGNAL_TRACK_COMPLETED :
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())
        }

    def __init__(self, profile):
        gobject.GObject.__init__(self)
        self._filesrc = self.build_filesrc()
        self._encoder = self.build_encoder(profile.pipeline)
        self._filesink = self.build_filesink()
        self._gst_pipe = None
        return

    def get_position(self):
        if self._gst_pipe is None:
            return None
        
        pos = self._gst_pipe.query_position(gst.FORMAT_TIME)[0]
        dur = self._gst_pipe.query_duration(gst.FORMAT_TIME)[0]
        per = float(pos)/float(dur) * 100
        return int(per)

    def start(self):
        self._gst_pipe = self.get_gst_pipeline()
        s = self._gst_pipe.set_state(gst.STATE_NULL)
        s = self._gst_pipe.set_state(gst.STATE_PLAYING)
        return

    def get_gst_pipeline(self):
        pipe = gst.Pipeline("cpipeline")
        pipe.add(self._filesrc, self._encoder, self._filesink)
        self._filesrc.link(self._encoder)
        self._encoder.link(self._filesink)
        bus = pipe.get_bus()
        bus.add_signal_watch()
        bus.connect("message::eos", self.eos_callback)
        bus.connect("message::error", self.error_callback)
        return pipe

    def set_sink_filename(self, name):
        self._filesink.set_state(gst.STATE_NULL)
        self._filesink.set_property("location", name)
        return

    def get_sink_filename(self):
        return self._filesink.get_property("location")

    sink_filename = property(get_sink_filename, set_sink_filename, None,
                             "The sink filename")

    def set_source_filename(self, name):
        self._filesrc.set_state(gst.STATE_NULL)
        self._filesrc.set_property("location", name)
        return

    def get_source_filename(self):
        return self._filesrc.get_property("location")

    source_filename = property(get_source_filename, set_source_filename, None,
                               "The source filename")

    def build_filesrc(self):
        return gst.element_factory_make("filesrc", "input")

    def connect_dbin(self, dbin, pad, is_last, audioconvert):
        if not pad.is_linked():
            if not "audio" in pad.get_caps().to_string():
                return
            pad.link(audioconvert.get_pad("sink"))
        return

    def build_encoder(self, pipestring):
        p = ("decodebin name=decode ! audioconvert name=convert ! %s" % 
             pipestring)
        print "encoder pipe:", p
        bin = gst.parse_launch("bin.( %s )" % p)

        dbin = bin.get_by_name("decode")
        audioconvert = bin.get_by_name("convert")
        enc = bin.get_by_name("enc")
    
        assert dbin
        assert audioconvert
        assert enc

        dbin.connect("new-decoded-pad", self.connect_dbin, audioconvert)

        # what we really want is the last element's src pad?
        bin.add_pad(gst.GhostPad("src", bin.find_unconnected_pad("src")))
        bin.add_pad(gst.GhostPad("sink", dbin.get_pad("sink")))
        return bin

#        pad = bin.find_unconnected_pad(gst.PAD_SRC)
#        if pad is not None:
#            bin.add_pad(gst.GhostPad("src", pad))
#        pad = bin.find_unconnected_pad(gst.PAD_SINK)
#        if pad is not None:
#            bin.add_pad(gst.GhostPad("sink", pad))
#        return bin

    def build_filesink(self):
        return gst.element_factory_make("filesink", "output")

    def eos_callback(self, bus, message, *data):
        self.emit(self.SIGNAL_TRACK_COMPLETED)
        return

    def error_callback(self, bus, message, *data):
        error, debug = message.parse_error()
        print >>sys.stderr,"Error in conversion:"
        print >>sys.stderr,"  ", error
        print >>sys.stderr,"  ", debug
        return

class FilenameException(Exception):
    pass

class Track(gobject.GObject, object):
    SIGNAL_TRACK_COMPLETED = "completed"

    __slots__ = ('_source', '_target', '_target_profile', '_pipeline')

    __gsignals__ = {
        SIGNAL_TRACK_COMPLETED :
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())
        }

    def __init__(self, source, target_profile, target_directory):
        gobject.GObject.__init__(self)
        self._source = source
        try:
            os.makedirs(target_directory)
        except OSError, er:
            if er.errno != 17: # ignore file exists
                raise
        self._target = self.create_target_filename(
            source, target_profile, target_directory)
        self._target_profile = target_profile
        self._pipeline = Pipeline(target_profile)
        self._complete = False
        return

    complete = property(lambda self: self._complete, None, None, 
                        "Is the conversion complete")

    def get_position(self):
        return self._pipeline.get_position()

    def convert(self):
        self._pipeline.connect(
            Pipeline.SIGNAL_TRACK_COMPLETED, self.completed_cb)
        self._pipeline.source_filename = self.source
        self._pipeline.sink_filename = self.target
        self._pipeline.start()
        return

    def completed_cb(self, *args):
        self._complete = True
        self.emit(self.SIGNAL_TRACK_COMPLETED)
        return

    def get_source(self):
        return self._source

    source = property(get_source, None, None, "The source track")

    def get_target(self):
        return self._target

    target = property(get_target, None, None, "The target track")

    # private

    def create_target_filename(self, source, target_profile, target_directory):
        def dtarget(n):
            return os.path.join(target_directory, n)
        target_name = self.get_unix_safe_target(
            self.get_format_filename(
                target_profile,
                os.path.basename(source)))
        if os.path.exists(dtarget(target_name)):
            increment = 2
            incremented_name = target_name
            while os.path.exists(dtarget(incremented_name)):
                incremented_name = self.increment_filename(
                    target_name, increment)
                increment = increment + 1
            target_name = incremented_name
        if self.can_create_file(dtarget(target_name)):
            return dtarget(target_name)
        target_name = self.get_fat_safe_target(target_name)
        if can_create_file(dtarget(target_name)):
            return dtarget(target_name)
        raise FilenameException("Can't create an acceptable version of %s in %s" % 
                                source, target_directory)

    def can_create_file(self, name):
        file = None
        try:
            try:
                file = open(name, "w")
                file.close()
            except Exception, ex:
                return False
        finally:
            try:
                os.unlink(name)
            except Exception, ex:
                pass
        return True

    def get_format_filename(self, profile, filename):
        return ".".join((filename.rsplit(".", 1)[0], profile.extension))

    UNIX_UNSAFE_CHARS_RE = re.compile(r"/")
    FAT_UNSAFE_CHARS_RE = re.compile(r"[:\\]")

    def increment_filename(self, name, increment):
        head, extension = name.rsplit(".", 1)
        return "".join([head, " (", str(increment), ").", extension])

    def get_unix_safe_target(self, name):
        return self.UNIX_UNSAFE_CHARS_RE.sub("-", name)

    def get_fat_safe_target(self, name):
        return self.FAT_UNSAFE_CHARS_RE.sub("-", name)

def start(start_mainloop):
    files = parse_arguments()
    ac = AudioConverter()
    ac.track_names = files
    acui = AudioConverterDialog(ac, start_mainloop)
    if start_mainloop:
        gtk.main()

def parse_arguments():
    parser = optparse.OptionParser(
        usage="%s [options] files" % NAME,
        version="%s %s" % (NAME, VERSION))
    options, args = parser.parse_args()
    return args

if HAS_NAUTILUS:
    class AudioFormatExtension(nautilus.MenuProvider):
        def __init__(self):
            f = open("/tmp/audioformat.error", "w", 0)
            sys.stderr = f
            print >>sys.stderr, "AudioFormat error log"
            pass

        def menu_activate_cb(self, menu, files):
            start([urllib.unquote(f.get_uri()[7:]) 
                   for f in files if not f.is_gone()], False)
            return

        def get_file_items(self, window, files):
            for file in files:
                mime = file.get_mime_type()
                if not mime.startswith("audio/") or mime == "application/ogg":
                    return
                if file.get_uri_scheme() != 'file':
                    return

            item = nautilus.MenuItem("Nautilus::AudioFormat",
                                     "Format audio...",
                                     "Convert to different audio formats")
            item.connect('activate', self.menu_activate_cb, files)
            return item,

if __name__ == '__main__':
    start(True)

